from tiktoken import model, encoding_for_model
from math import ceil
from typing import NewType

"""
type aliasing for convenince
"""


def get_models():
    return [key for key in model.MODEL_TO_ENCODING.keys()]


Model = NewType("Model", str)
GPT_4O = Model("gpt-4o")
MODEL_CHOICES: set[Model] = set(Model(model) for model in get_models())


"""
Tokenizing methods
"""


def calculate_text_tokens(string: str, model_name: str) -> int:
    """Returns the number of tokens in a text string"""
    encoding = encoding_for_model(model_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens


# TODO: add switch statement for strategies old and new
# def calculate_image_tokens(width: int, height: int, model_name: str="TODO: fix me") -> int:
#     logging.debug(f"calculating tokens for {model_name}")
#     return old_calculate_image_tokens(width,height)
#
# def new_calculate_image_tokens(width: int, height: int) -> int:
#     logging.debug(f"calculating tokens for {width} x {height}")
#     raise NotImplementedError

# Source for original image token code: https://medium.com/@teekaifeng/gpt4o-visual-tokenizer-an-illustration-c69695dd4a39


# TODO: update with more up to date checks for models
# def old_calculate_image_tokens(width: int, height: int) -> int:
def calculate_image_tokens(width: int, height: int) -> int:
    """
    Returns the number of tokens in an image
    """
    # below is only valid for the following models:
    # 4o, 4.1, 4.5
    # Step 1: scale to fit within a 2048 x 2048 square (maintain aspect ratio)
    if width > 2048 or height > 2048:
        aspect_ratio = width / height
        if aspect_ratio > 1:
            width, height = 2048, int(2048 / aspect_ratio)
        else:
            width, height = int(2048 * aspect_ratio), 2048

    # Step 2: scale such that the shortest side of the image is 768px long
    if width >= height and height > 768:
        width, height = int((768 / height) * width), 768
    elif height > width and width > 768:
        width, height = 768, int((768 / width) * height)

    # Step 3: compute number of 512x512 tiles that can fit into the image
    tiles_width = ceil(width / 512)
    tiles_height = ceil(height / 512)

    # See https://platform.openai.com/docs/guides/images-vision?api-mode=chat#gpt-4o-gpt-4-1-gpt-4o-mini-cua-and-o-series-except-o4-mini
    #   - 85 is the "base token" that will always be added
    #   - 1 tiles = 170 tokens
    total_tokens = 85 + 170 * (tiles_width * tiles_height)

    return total_tokens



import logging
from . import cli, config, models

"""
Main function
"""


def main() -> None:
    p = cli.setup_argparse()
    args = p.parse_args()
    cfg = config.generate_config(vars(args))

    token_counter = models.TokenCounter(cfg)
    token_counter.add_exclusions(cfg.exclude)
    token_counter.add_inclusions(cfg.include)

    logging.debug("Parsing files...")

    token_counter.parse_files()

    logging.debug("Parsing complete!")

    token_counter.output()


if __name__ == "__main__":
    main()



from ddt import tokenizer
from typing import TypedDict


# TODO: update these two tests - they ONLY work for the default model of gpt-4o
def test_calclate_known_text_tokens():
    # TokenPairs exists to prevent pyright from throwing a type error
    class TokenPairs(TypedDict):
        text: str
        total: int

    KNOWN_TEXT_TOKENS: list[TokenPairs] = [
        {
            "text": "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
            "total": 22,
        },
        {
            "text": "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur.",
            "total": 41,
        },
        {
            "text": "Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum. Nunc accumsan semper libero quis vel nulla sagittis eget volutpat.",
            "total": 39,
        },
        {
            "text": "Nulla facilisi. Sed sit amet nulla auctor, vestibulum magna sed, convallis ex. Fusce at purus ut orci gravida sodales.",
            "total": 29,
        },
        {
            "text": "Nunc sed ante non metus lacinia finibus. Nam vel semper sapien, eu tempus libero. Aenean sit amet risus sit amet nisi fermentum iaculis.",
            "total": 35,
        },
        {
            "text": "Cras ultricies ligula sed magna dictum porta. Morbi eget erat a est suscipit egestas. Etiam non lectus vel ex pulvinar dignissim.",
            "total": 30,
        },
        {
            "text": "Nullam accumsan nibh ut arcu sodales, id luctus dolor facilisis. Pellentesque in neque et leo pharetra luctus sit amet at libero.",
            "total": 31,
        },
    ]

    for case in KNOWN_TEXT_TOKENS:
        assert tokenizer.calculate_text_tokens(case["text"], "gpt-4o") == case["total"]


def test_calculate_known_image_tokens():
    KNOWN_IMAGE_TOKENS = [
        {"width": 1024, "height": 1024, "total": 765},
        {"width": 2048, "height": 4096, "total": 1105},
        {"width": 4096, "height": 2048, "total": 1105},
    ]

    for case in KNOWN_IMAGE_TOKENS:
        assert (
            tokenizer.calculate_image_tokens(case["width"], case["height"])
            == case["total"]
        )



from pathlib import Path
import sys
from ddt import config, tokenizer


def test_config_happy_path():
    root = Path(".")
    is_verbose = True
    include_gitignore = True
    include_dotfiles = True
    include_symlinks = True
    include_images = True
    resolve_paths = True
    model = tokenizer.Model("gpt-4o")
    output = sys.stdout
    output_format = "txt"
    exclude = ["foo"]
    include = ["bar"]

    cfg = config.Config(
        root,
        is_verbose,
        include_gitignore,
        include_dotfiles,
        include_symlinks,
        include_images,
        resolve_paths,
        model,
        output,
        output_format,
        exclude,
        include,
    )

    assert cfg.root == root
    assert cfg.is_verbose
    assert cfg.include_gitignore
    assert cfg.include_dotfiles
    assert cfg.include_symlinks
    assert cfg.include_images
    assert cfg.resolve_paths
    assert cfg.model == tokenizer.Model("gpt-4o")
    assert cfg.output == sys.stdout
    assert cfg.output_format == "txt"
    assert cfg.exclude == ["foo"]
    assert cfg.include == ["bar"]


def test_generate_config():
    args = [
        {
            "root": Path("."),
            "verbose": True,
            "include_gitignore": True,
            "include_dotfiles": True,
            "include_symlinks": True,
            "include_images": True,
            "resolve_paths": True,
            "model": tokenizer.Model("gpt-4o"),
            "output": sys.stdout,
            "output_type": "txt",
            "json": False,  # to fix test, goes away when output config changes
            "html": False,  # to fix test, goes away when output config changes
            "exclude": ["foo"],
            "include": ["bar"],
        },
        {
            "root": Path("."),
            "verbose": True,
            "include_gitignore": True,
            "include_dotfiles": True,
            "include_symlinks": True,
            "include_images": True,
            "resolve_paths": True,
            "model": tokenizer.Model("gpt-4o"),
            "output": sys.stdout,
            "output_type": "json",
            "json": True,  # to fix test, goes away when output config changes
            "html": False,  # to fix test, goes away when output config changes
            "exclude": ["foo"],
            "include": ["bar"],
        },
        {
            "root": Path("."),
            "verbose": True,
            "include_gitignore": True,
            "include_dotfiles": True,
            "include_symlinks": True,
            "include_images": True,
            "resolve_paths": True,
            "model": tokenizer.Model("gpt-4o"),
            "output": sys.stdout,
            "output_type": "html",
            "json": False,  # to fix test, goes away when output config changes
            "html": True,  # to fix test, goes away when output config changes
            "exclude": ["foo"],
            "include": ["bar"],
        },
    ]

    for argset in args:
        cfg = config.generate_config(argset)

        assert cfg.root == Path(".").resolve()
        assert cfg.is_verbose
        assert cfg.include_gitignore
        assert cfg.include_dotfiles
        assert cfg.include_symlinks
        assert cfg.include_images
        assert cfg.resolve_paths
        assert cfg.model == tokenizer.Model("gpt-4o")
        assert cfg.output == sys.stdout
        assert cfg.output_format == argset["output_type"]
        assert cfg.exclude == ["foo"]
        assert cfg.include == ["bar"]



import mimetypes
import logging
from pathlib import Path
from typing import Any
from typing import override as orr
from . import tokenizer
from .config import Config
from PIL import Image
from jinja2 import Environment, PackageLoader, select_autoescape
import json

"""
Data models
"""


class TokenCounter:
    """
    A class representing the contents of a directory and the count of tokens per file in that directory.

    Attributes:
        root(Path): The root path of the directory.
        all_files(list[Path]): All file paths in the directory.
        ignored_files(dict[str, list[Path]]): All files ignored by the scan, grouped by extension.
        scanned_files(dict[str, FileCategory]): All files scanned, grouped by extension.
        excluded_files(set[Path]): All files to be excluded from the search.
        included_files(set[Path]): All files to be included in the search.
        total(int): The total number of tokens present within the directory.
        config(Config): the configuration file for the TokenCounter being run.
    """

    def __init__(self, cfg: Config) -> None:
        mimetypes.init()
        self.config: Config = cfg
        self.all_files: list[Path] = (
            [file.resolve() for file in self.config.root.glob("**/*.*")]
            if self.config.resolve_paths
            else [file for file in self.config.root.glob("**/*.*")]
        )
        self.ignored_files: dict[str, list[Path]] = {}
        self.scanned_files: dict[str, FileCategory] = {}
        self.excluded_files: set[Path] = set()
        self.included_files: set[Path] = set()
        self.total: int = 0

    def to_dict(self) -> dict[str, Any]:
        """
        Converts TokenCounter to a dictionary type for JSON encoding.
        """
        return {
            "root": str(self.config.root),
            "all_files": [path.name for path in self.all_files],
            "ignored_files": {
                key: [path.name for path in paths]
                for key, paths in self.ignored_files.items()
            },
            "scanned_files": {
                ext: category.to_dict() for ext, category in self.scanned_files.items()
            },
            "total": self.total,
        }

    def to_text(self) -> str:
        """
        Converts TokenCounter to an ASCII-style table.
        """
        result: str = ""
        if self.config.is_verbose:
            result += "ignored:\n"
            for extension, ignored in self.ignored_files.items():
                result += "=========================\n"
                result += f"{extension} files ignored:\n"
                result += "*************************\n"
                for file in ignored:
                    result += f"{str(file)}\n"
            result += "=========================\n"
        result += "totals:\n"
        for extension, file_extension in self.scanned_files.items():
            result += "-------------------------\n"
            result += f"{extension} tokens:\n"
            result += "*************************\n"
            for file in file_extension.files:
                result += f"{file['file']}: {file['tokens']:,} tokens\n"
            result += ".........................\n"
            result += (
                f"{file_extension.extension} total: {file_extension.total:,} tokens\n"
            )

        result += "-------------------------\n"
        result += f"grand total: {self.total:,}\n"
        result += (
            f"remaining tokens given 128K context window: {128_000 - self.total:,}\n"
        )
        return result

    def to_html(self) -> str:
        """
        Converts TokenCounter to an HTML table.
        """
        env = Environment(loader=PackageLoader("ddt"), autoescape=select_autoescape())
        template = env.get_template("template.html")
        values: dict[
            str, Path | bool | dict[str, list[Path]] | dict[str, FileCategory] | int
        ] = {
            "directory": self.config.root,
            "verbose": self.config.is_verbose,
            "ignored_files": self.ignored_files,
            "scanned_files": self.scanned_files,
            "total": self.total,
        }
        result: str = template.render(values)
        return result

    def add_exclusions(self, exclusions: list[str]) -> None:
        """
        Adds filetypes to the excluded files list.

        Args:
            exclusions (list[str]): The file extensions to be excluded.
        """
        if exclusions is None or len(exclusions) < 1:
            return
        for ext in exclusions:
            for file in self.config.root.glob(f"**/*.{ext}"):
                self.excluded_files.add(file.resolve())

    def add_inclusions(self, inclusions: list[str]) -> None:
        """
        Adds files to the included files list.

        Args:
            inclusions (list[str]): The file extensions to be included.
        """
        if inclusions is None or len(inclusions) < 1:
            return
        for ext in inclusions:
            for file in self.config.root.glob(f"**/*.{ext}"):
                self.included_files.add(file.resolve())

    def count_text_file(self, file: Path) -> int:
        """
        Parses the given text file and return the total number of tokens.

        Args:
            file (Path): The text file being processed
            file_extension (str): The suffix of the filetype.

        Returns:
            int: the total number of tokens in the file.
        """
        try:
            text = file.read_text()
        except UnicodeDecodeError:
            logging.debug(f"file {file.name} hit unicode error, ignoring")
            self.add_to_ignored(file)
            return 0
        return tokenizer.calculate_text_tokens(text, self.config.model)

    def count_image_file(self, file: Path) -> int:
        """
        Parses the given image file and return the total number of tokens.

        Args:
            file (Path): The text image being processed
            file_extension (str): The suffix of the filetype.

        Returns:
            int: the total number of tokens in the file.
        """
        try:
            img = Image.open(file)
            width, height = img.size
            return tokenizer.calculate_image_tokens(width, height)
        except Exception as e:
            logging.debug(f"file {file.name} hit error {e}, ignoring")
            self.add_to_ignored(file)
            return 0

    def add_to_ignored(self, file: Path):
        """
        Adds the given file to the ignored files list.
        """
        filetype = self.grab_suffix(file)
        logging.debug(f"ignoring {str(file)}")
        if filetype not in self.ignored_files:
            self.ignored_files[filetype] = []
        self.ignored_files[filetype].append(file)

    def filter_file(self, file: Path) -> bool:
        if len(self.included_files) > 0 and file not in self.included_files:
            self.add_to_ignored(file)
            return True

        if len(self.excluded_files) > 0 and file in self.excluded_files:
            self.add_to_ignored(file)
            return True

        if not self.config.include_dotfiles and any(
            part.startswith(".") for part in file.parts
        ):
            self.add_to_ignored(file)
            return True

        if not self.config.include_gitignore and file in self.config.gitignore:
            self.add_to_ignored(file)
            return True

        if (
            not self.config.include_symlinks
            and self.config.root.name not in file.resolve().parts
        ):
            self.add_to_ignored(file)
            return True

        return False

    def parse_file(self, file: Path) -> tuple[str, int]:
        """
        Parses an individual file. Checks the MIME, and
        """
        file_extension = self.grab_suffix(file)

        # TODO: start subdividing by mimetypes and set up your own mimetype list
        # file_type = mimetypes.guess_file_type(file)
        # logging.debug(f"filetype guess: {file_type}")
        mime: str | None = (
            mimetypes.types_map[file_extension]
            if file_extension in mimetypes.types_map
            else None
        )

        logging.debug(f"reading {str(file)}")

        if mime:
            category = mime.split("/")[0]
            match category:
                case "image":
                    if self.config.include_images:
                        token_counts = self.count_image_file(file)
                    else:
                        self.add_to_ignored(file)
                        return "", 0
                case _:
                    # currently assuming everything is a text file if it's not an image
                    token_counts = self.count_text_file(file)
        else:
            token_counts = self.count_text_file(file)
        return file_extension, token_counts

    def parse_files(self):
        """
        Iterate through `self.all_files`, skip uncounted files, and parse the rest.
        Anything that didn't result in a token count is skipeped, and the rest are added
        to the list of scanned files.

        """
        for file in self.all_files:
            logging.debug(f"checking {str(file)}")
            if file.is_dir():
                continue

            filtered: bool = self.filter_file(file)
            if filtered:
                continue

            file_extension, token_counts = self.parse_file(file)

            # TODO: handle the case of empty files - right now we're treating empty and error as the same
            if token_counts == 0:
                continue

            if file_extension not in self.scanned_files:
                self.scanned_files[file_extension] = FileCategory(file_extension)
            self.scanned_files[file_extension].files.append(
                {"file": file.name, "tokens": token_counts}
            )
            self.scanned_files[file_extension].total += token_counts
            self.total += token_counts

    def grab_suffix(self, file: Path) -> str:
        """
        A helper module to handle the cases where a filetype has multiple periods in the extension, e.g. .tar.gz

        Args:
            file (Path): The file path

        Returns:
            str: the full file path.
        """
        if len(file.suffixes) == 1:
            return file.suffix
        result = ""
        for suffix in file.suffixes:
            result += suffix
        return result

    def output(self) -> None:
        """
        Writes the contents of the tokenencoder to the eoutput.
        """
        if not self.config.output:
            exit(1)
        with self.config.output as f:
            match self.config.output_format:
                case "json":
                    json.dump(self, f, cls=TokenCounterEncoder, indent=2)
                case "html":
                    _ = f.write(self.to_html())
                case _:
                    _ = f.write(self.to_text())


class TokenCounterEncoder(json.JSONEncoder):
    """
    A custom token encoder that overrides the default() method to allow encoding of the TokenCounter to JSON
    """

    # was having some issues with the @override decorator, this is a workaround
    @orr
    def default(self, o: Any) -> Any:
        if isinstance(o, TokenCounter) or isinstance(o, FileCategory):
            return o.to_dict()
        return super().default(o)


class FileCategory:
    """
    A class representing a particular type of file, like a .txt or .yaml.

    Args:
    extension: str - The file extension, e.g. .txt

    Attributes:
    extension: str - The file extension, e.g. .txt
    files: list[dict[str, str | int]] - The files, with the structure {"file": str, "tokens": int}
    tital: int - the total number of tokens in this file category.

    """

    def __init__(self, extension: str) -> None:
        self.extension: str = extension
        self.files: list[dict[str, str | int]] = []
        self.total: int = 0

    def to_dict(self):
        """
        Converts TokenCounter to a dictionary type for JSON encoding.
        """
        return {
            "total": self.total,
            "files": self.files,
        }



import logging
from pathlib import Path
from typing import Any, TextIO
from dataclasses import dataclass, field
from . import tokenizer

"""
Config model
"""


@dataclass
class Config:
    """
    A class representing the configuration for the current run of DDT.

    Properties:
        root (Path): The starting directory for the path traversal.
        is_verbose (bool): Verbosity flag - True logs DEBUG output.
        include_gitignore (bool): Flag - True counts tokens of gitignored files.
        include_dotfiles (bool): Flag - True counts tokens of dotfiles and dotfile directories.
        include_symlinks (bool): Flag - True counts tokens of symbolically linked files.
        include_images (bool): Flag - True counts image tokens.
        resolve_paths (bool): Flag - True displays file names by their absolute path.
        model (Model): The specified model for the encoding algorithms.
        output (TextIO): The output stream for results.
        output_format (str): The output encoding format.
        exclude (list[str]): The list of user-specified filetypes to exclude
        include (list[str]): The list of user-specified filetypes to include - all other types are ignored.
        gitignore (set[Path]): The files found within the gitignore.
    """

    root: Path
    is_verbose: bool
    include_gitignore: bool
    include_dotfiles: bool
    include_symlinks: bool
    include_images: bool
    resolve_paths: bool
    model: tokenizer.Model
    output: TextIO
    output_format: str
    exclude: list[str]
    include: list[str]
    gitignore: set[Path] = field(init=False)

    def __post_init__(self):
        self._setup_logging()
        self.gitignore = self._parse_gitignore()

    # AI rewrote this function for me, need to replace.
    def _parse_gitignore(self) -> set[Path]:
        # TODO: ensure that this matches the gitignore spec: https://git-scm.com/docs/gitignore
        """
        Reads the .gitignore file in the given root directory, interprets its patterns,
        and returns a set of Paths representing all files and directories within root that match
        those patterns. Lines that are empty or start with '#' (comments) are ignored.

        For pattern matching:
          - Patterns that start with '/' are treated as relative to the root.
          - Patterns that contain a slash (but do not start with '/') are also treated as relative.
          - Patterns without any slash are searched recursively using rglob.
          - If a pattern ends with '/', it is interpreted as a directory (the trailing slash is removed
            before matching).

        Args:
            root (Path): The root directory containing the .gitignore file.

        Returns:
            Set[Path]: A set of Paths that match the patterns specified in the .gitignore file.
        """

        if not self.root.is_dir():
            logging.info(f"{self.root} is not a directory, exiting.")
            exit(1)

        ignored: set[Path] = set()
        gitignore_file: Path = self.root / ".gitignore"

        try:
            with gitignore_file.open("r") as f:
                patterns: list[str] = []
                for line in f:
                    stripped = line.strip()
                    # Skip empty lines or comments.
                    if not stripped or stripped.startswith("#"):
                        continue
                    patterns.append(stripped)
        except FileNotFoundError:
            return ignored  # No .gitignore file found.

        for pattern in patterns:
            # Check if the pattern is meant for directories (ends with a slash)
            if pattern.endswith("/"):
                # Remove trailing slash for glob matching.
                pattern = pattern.rstrip("/")

            # If the pattern starts with '/', it is anchored to the root.
            if pattern.startswith("/"):
                # Remove the leading slash.
                pattern = pattern[1:]
                # Use glob relative to the root (non-recursive).
                matches = self.root.glob(pattern)
            # If the pattern contains a slash somewhere, treat it as relative to the root.
            elif "/" in pattern:
                matches = self.root.glob(pattern)
            else:
                # Pattern without a slash: search recursively.
                matches = self.root.rglob(pattern)

            for match in matches:
                ignored.add(Path(match))

        return ignored

    def _setup_logging(self) -> None:
        level = logging.DEBUG if self.is_verbose else logging.INFO
        logging.basicConfig(format="%(message)s", level=level)


def generate_config(args: dict[str, Any]) -> Config:
    """
    Basically the main function of this class - it converts the CLI config into a system config:

    - Fills out config with existing config file.
    - Specifies output format.
    - applies CLI args to config
    - Resolves the root path and validates that it is a directory
    - Generates and returns a config based on the CLI and JSON config
    """
    if args["json"]:
        output_format = "json"
    elif args["html"]:
        output_format = "html"
    else:
        output_format = "txt"

    cfg = Config(
        root=args["root"].resolve(),
        is_verbose=args["verbose"],
        include_gitignore=args["include_gitignore"],
        include_dotfiles=args["include_dotfiles"],
        include_symlinks=args["include_symlinks"],
        include_images=args["include_images"],
        resolve_paths=args["resolve_paths"],
        model=args["model"],
        output=args["output"],
        output_format=output_format,
        exclude=args["exclude"],
        include=args["include"],
    )
    return cfg



import os
import pytest
from pathlib import Path
from ddt import cli, tokenizer


def test_root_arg():
    parser = cli.setup_argparse()
    args = parser.parse_args(["src"])
    assert args.root == Path("src")
    with pytest.raises(SystemExit):
        _ = parser.parse_args([])


def test_verbosity_arg():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--verbose", "src"])
    assert args.verbose


def test_includes_args():
    parser = cli.setup_argparse()
    args = parser.parse_args(
        [
            "--include-gitignore",
            "--include-dotfiles",
            "--include-symlinks",
            "--include-images",
            "src",
        ]
    )
    assert args.include_gitignore
    assert args.include_dotfiles
    assert args.include_symlinks
    assert args.include_images


def test_resolve_paths_arg():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--resolve-paths", "src"])
    assert args.resolve_paths


def test_model_choices_arg():
    parser = cli.setup_argparse()
    for model in tokenizer.MODEL_CHOICES:
        args = parser.parse_args(["--model", model, "src"])
        assert args.model == model
    with pytest.raises(SystemExit):
        _ = parser.parse_args([".", "-m", "invalid_model"])


def test_output_arg():
    parser = cli.setup_argparse()
    # TODO: replace with pytest temp dir stuff
    with open("test.txt", "w") as file:
        args = parser.parse_args(["--output", "test.txt", "src"])
        assert args.output.name == file.name
        length = args.output.write("foo")
        assert length == 3
        os.remove(file.name)
    with pytest.raises(SystemExit):
        _ = parser.parse_args([".", "-o", "src/"])


def test_output_type_args():
    parser = cli.setup_argparse()
    with pytest.raises(SystemExit):
        _ = parser.parse_args([".", "--json", "--html"])


def test_output_json_arg():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--json", "src"])
    assert args.json


def test_output_html_arg():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--html", "src"])
    assert args.html


def test_exclude_args():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--exclude", ".py", "src"])
    assert args.exclude == [".py"]


def test_include_args():
    parser = cli.setup_argparse()
    args = parser.parse_args(["--include", ".py", "src"])
    assert args.include == [".py"]


def test_input_filter_args():
    parser = cli.setup_argparse()
    with pytest.raises(SystemExit):
        _ = parser.parse_args([".", "--include", ".py", "--exclude", ".json"])



import mimetypes
from pathlib import Path
from .config import PathConfig

"""
One data stucture:
All Files - a set of files (Path)
Excluded files - a list/set of files (Path)
Included files - a list/set of files (Path)

Logic:
take a FileTree.All Files and filter it.

Input:

- Set of files

Logic:

1. filter excluded files to get initial included/excluded list, and add exclusions to Results.NotCounted
2. if custom inclusions exist, filter included files to get added exclusions and new included list. Add new exclusions to Results.NotCounted
4. Attempt to load included files to memory:
    4a. If successful, load into TokenQueue
    4b. If unsuccessful, add to Results.NotCounted
"""


def filter_file(file: Path, config: PathConfig, gitignore: set[Path]) -> bool:
    mime_type, _ = mimetypes.guess_type(file)
    if not mime_type:
        return True
    base_type = mime_type.split("/")[0]
    if base_type == "video":
        return True
    if base_type == "image" and not config.include_images:
        return True
    if not config.include_symlinks and file.is_symlink():
        return True
    if not config.include_dotfiles and any(part.startswith(".") for part in file.parts):
        return True
    if not config.include_gitignore and file in gitignore:
        return True
    if excluded_file(file, config.exclude):
        return True
    if not included_file(file, config.include):
        return True
    return False


def excluded_file(file: Path, exclusions: list[str]):
    return len(exclusions) > 0 and file in exclusions


def included_file(file: Path, inclusions: list[str]):
    return len(inclusions) > 0 and file in inclusions


def parse_gitignore(directory: Path) -> set[Path]:
    # TODO: ensure that this matches the gitignore spec: https://git-scm.com/docs/gitignore
    """
    Reads the .gitignore file in the given root directory, interprets its patterns,
    and returns a set of Paths representing all files and directories within root that match
    those patterns. Lines that are empty or start with '#' (comments) are ignored.

    For pattern matching:
      - Patterns that start with '/' are treated as relative to the root.
      - Patterns that contain a slash (but do not start with '/') are also treated as relative.
      - Patterns without any slash are searched recursively using rglob.
      - If a pattern ends with '/', it is interpreted as a directory (the trailing slash is removed
        before matching).

    Args:
        root (Path): The root directory containing the .gitignore file.

    Returns:
        Set[Path]: A set of Paths that match the patterns specified in the .gitignore file.
    """

    ignored: set[Path] = set()
    gitignore_file: Path = directory / ".gitignore"

    try:
        with gitignore_file.open("r") as f:
            patterns: list[str] = []
            for line in f:
                stripped = line.strip()
                # Skip empty lines or comments.
                if not stripped or stripped.startswith("#"):
                    continue
                patterns.append(stripped)
    except FileNotFoundError:
        return ignored  # No .gitignore file found.

    for pattern in patterns:
        # Check if the pattern is meant for directories (ends with a slash)
        if pattern.endswith("/"):
            # Remove trailing slash for glob matching.
            pattern = pattern.rstrip("/")

        # If the pattern starts with '/', it is anchored to the root.
        if pattern.startswith("/"):
            # Remove the leading slash.
            pattern = pattern[1:]
            # Use glob relative to the root (non-recursive).
            matches = directory.glob(pattern)
        # If the pattern contains a slash somewhere, treat it as relative to the root.
        elif "/" in pattern:
            matches = directory.glob(pattern)
        else:
            # Pattern without a slash: search recursively.
            matches = directory.rglob(pattern)

        for match in matches:
            ignored.add(Path(match))

    return ignored



from pathlib import Path
import argparse
import sys
from . import tokenizer


def setup_argparse() -> argparse.ArgumentParser:
    """
    Configures the CLI flags.
    """
    parser = argparse.ArgumentParser(
        prog="Tokenizer",
        description="Crawls a given directory, counts the number of tokens per filetype in the project and returns a per-type total and grand total",
        epilog="Made with <3 by 0x4D5352",
    )

    _ = parser.add_argument(
        "root",
        help="the relative or absolute path to the directory you wish to scan",
        type=Path,
    )

    _ = parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="set to increase logging to console",
    )

    _ = parser.add_argument(
        "-g",
        "--include-gitignore",
        action="store_true",
        help="include files and directories found in the .gitignore file",
    )
    _ = parser.add_argument(
        "-d",
        "--include-dotfiles",
        action="store_true",
        help="include files and directories beginning with a dot (.)",
    )
    _ = parser.add_argument(
        "-s",
        "--include-symlinks",
        action="store_true",
        help="include files and directories symlinked from outside the target directory",
    )
    _ = parser.add_argument(
        "-i",
        "--include-images",
        action="store_true",
        help="include image files found within the directory",
    )

    _ = parser.add_argument(
        "-r",
        "--resolve-paths",
        action="store_true",
        help="resolve relative file paths to their absolute location",
    )

    _ = parser.add_argument(
        "-m",
        "--model",
        action="store",
        help="specify a model to use for token approximation. default is 'gpt-4o'",
        choices=tokenizer.MODEL_CHOICES,
        default=tokenizer.GPT_4O,
        type=tokenizer.Model,
    )

    _ = parser.add_argument(
        "-o",
        "--output",
        action="store",
        help="redirect output from STDOUT to a file at the location specified.",
        type=argparse.FileType(mode="w", encoding="UTF-8"),
        default=sys.stdout,
    )

    output_type_group = parser.add_mutually_exclusive_group()
    _ = output_type_group.add_argument(
        "--json",
        action="store_true",
        help="save the results of the scan to a json file",
        deprecated=True,
    )
    _ = output_type_group.add_argument(
        "--html",
        action="store_true",
        help="save the results of the scan to a HTML file",
        deprecated=True,
    )

    input_filter_group = parser.add_mutually_exclusive_group()
    _ = input_filter_group.add_argument(
        "--exclude",
        action="append",
        help="specify file formats to ignore from counting. this flag may be set multiple times for multiple entries. cannot be set if including files",
        type=str,
    )
    _ = input_filter_group.add_argument(
        "--include",
        action="append",
        help="specify file formats to include when counting. this flag may be set multiple times for multiple entries. cannot be set if excluding files",
        type=str,
    )
    return parser



from dataclasses import dataclass
from pathlib import Path


@dataclass
class PathConfig:
    include_gitignore: bool
    include_dotfiles: bool
    include_symlinks: bool
    include_images: bool
    exclude: list[str]
    include: list[str]


def add_exclusions(directory: Path, exclusions: list[str]) -> list[str] | None:
    if exclusions is None or len(exclusions) < 1:
        return
    # TODO: make this account for files or filetypes (with or without a period)
    for ext in exclusions:
        for file in directory.glob(f"**/*.{ext}"):
            excluded_files.add(file.resolve())


def add_inclusions(self, inclusions: list[str]) -> None:
    if inclusions is None or len(inclusions) < 1:
        return
    for ext in inclusions:
        for file in self.config.root.glob(f"**/*.{ext}"):
            self.included_files.add(file.resolve())



"""
Two categories of Input:
    - Multiple Files:
        a. single directory
        b. list of directories
        c. list of files
    - Single File:
        a. single file
        b. std input


Logic:

0. Does root not exist but stdin does? If so, treat stdin as root.
1. Take `root` and try to convert it to a path.
2. Is it a resolvable path?
    2a. If yes, is it a directory?
        - If yes, glob and pass a set of all files to FileTree.All Files
        - If no, pass to TokenCounter.Parser
    2b. If no, is it json, csv, or text?
        - if JSON or CSV, repeat step 2 for each item
        - if text, split on space and repeat step 2 for each item
        - for both, count length before iterating and reject if over some arbitrary limit
        - for both, track Not A File count.
            - If exceeded, pass original results to TokenCounter.Parser
3. Take final list from 2b and pass a set of all files to FileTree.All Files
"""



import os
from pathlib import Path
import sys
from ddt import config, models
from ddt.tokenizer import Model

"""
    token counter
"""


def test_tokencounter_init():
    cfg = config.Config(
        Path("."),
        False,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    assert tc.config == cfg
    assert tc.all_files == [file for file in Path(".").glob("**/*.*")]
    assert tc.ignored_files == dict()
    assert tc.scanned_files == dict()
    assert tc.excluded_files == set()
    assert tc.included_files == set()
    assert tc.total == 0


def test_tokencounter_to_dict():
    cfg = config.Config(
        Path("."),
        False,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    result = tc.to_dict()
    assert result["root"] == str(Path("."))
    assert result["all_files"] == [file.name for file in Path(".").glob("**/*.*")]
    assert result["ignored_files"] == dict()
    assert result["scanned_files"] == dict()
    assert result["total"] == 0


def test_tokencounter_to_text():
    cfg = config.Config(
        Path("."),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    result = tc.to_text()
    assert "=========================\n" in result
    assert "ignored:" in result
    assert "total:" in result


def test_tokencounter_to_html():
    cfg = config.Config(
        Path("."),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    result = tc.to_html()
    assert "<!DOCTYPE html>" in result
    assert "Scanned files in" in result
    assert "<table>" in result
    assert "</table>" in result
    assert '<th scope="row">Grand Total</th>' in result


def test_tokencounter_add_exclusions():
    cfg = config.Config(
        Path("assets"),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    tc.add_exclusions([])
    assert tc.included_files == set()
    exclusions = ["tape"]
    tc.add_exclusions(exclusions)
    assert tc.excluded_files == {Path("assets/demo.tape").resolve()}


def test_tokencounter_add_inclusions():
    cfg = config.Config(
        Path("assets"),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    tc.add_inclusions([])
    assert tc.included_files == set()
    inclusions = ["tape"]
    tc.add_inclusions(inclusions)
    assert tc.included_files == {Path("assets/demo.tape").resolve()}


def test_tokencounter_count_text_file():
    cfg = config.Config(
        Path("."),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    count = tc.count_text_file(Path("assets/demo.gif"))
    assert count == 0
    count = tc.count_text_file(Path("assets/demo.tape"))
    assert count == 830


def test_tokencounter_count_image_file():
    cfg = config.Config(
        Path("."),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    count = tc.count_image_file(Path("assets/demo.tape"))
    assert count == 0
    count = tc.count_image_file(Path("assets/demo.gif"))
    assert count == 1105


def test_tokencounter_add_to_ignored():
    cfg = config.Config(
        Path("."),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    file = Path("README.md")
    tc.add_to_ignored(file)
    assert file in tc.ignored_files[".md"]


def test_tokencounter_filter_file():
    cfg = config.Config(
        Path("./tests/test_files"),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)

    inclusions = ["json"]
    tc.add_inclusions(inclusions)
    not_included = Path("output.json")
    inclusion_filtered = tc.filter_file(not_included)
    assert inclusion_filtered
    assert tc.ignored_files[".json"] == [not_included]
    tc.included_files = set()
    tc.ignored_files = dict()

    # TODO: figure out why this isn't getting flagged
    exclusions = ["html"]
    tc.add_exclusions(exclusions)
    excluded_file = Path("output.html")
    exclusion_filtered = tc.filter_file(excluded_file)
    assert exclusion_filtered
    assert tc.ignored_files[".html"] == [excluded_file]
    tc.excluded_files = set()
    tc.ignored_files = dict()

    # dotfiles
    dotfile = Path(".gitignore")
    dotfile_filtered = tc.filter_file(dotfile)
    assert dotfile_filtered
    assert [dotfile] in tc.ignored_files.values()
    tc.ignored_files = dict()

    # TODO: figure out why this isn't getting flagged
    # gitignore
    # TODO: replace with pytest temp dir stuff
    with open("gitignored.json", "w") as file:
        _ = file.write('{"foo": "bar"}')
    gitignored = Path("gitignored.json")
    gitignore_filtered = tc.filter_file(gitignored)
    assert gitignore_filtered
    assert tc.ignored_files[".json"] == [gitignored]
    tc.ignored_files = dict()
    os.remove(gitignored)

    # symlinks
    symlinked = Path("README.md")
    symlink_filtered = tc.filter_file(symlinked)
    assert symlink_filtered
    assert tc.ignored_files[".md"] == [symlinked]


def test_tokencounter_parse_file():
    cfg = config.Config(
        Path("tests/test_files"),
        True,
        False,
        False,
        False,
        True,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    file = Path("tests/test_files/testfile.txt")
    extension, token_count = tc.parse_file(file)
    assert extension == ".txt"
    assert token_count == 22
    image = Path("assets/contextwindow.png")
    extension, token_count = tc.parse_file(image)
    assert extension == ".png"
    assert token_count == 765


def test_tokencounter_parse_files():
    cfg = config.Config(
        Path("tests/test_files"),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    tc.parse_files()
    assert tc.ignored_files[".jpeg"] == [
        Path("tests/test_files/test_image.jpeg"),
    ]
    assert tc.ignored_files[""] == [
        Path("tests/test_files/.gitignore"),
        Path("tests/test_files/subtest/.invisible"),
    ]

    assert tc.scanned_files[".txt"].files == [
        {"file": Path("tests/test_files/testfile.txt").name, "tokens": 22},
    ]
    assert tc.total == 22


def test_tokencounter_grab_suffix():
    cfg = config.Config(
        Path("assets"),
        True,
        False,
        False,
        False,
        False,
        False,
        Model("gpt-4o"),
        sys.stdout,
        "txt",
        [],
        [],
    )
    tc = models.TokenCounter(cfg)
    file = Path("README.md")
    suffix = tc.grab_suffix(file)
    assert suffix == ".md"
    # TODO: replace with pytest temp dir stuff
    with open("foobar.tar.gz", "w") as f:
        _ = f.write("hello mom")
    file = Path("foobar.tar.gz")
    suffix = tc.grab_suffix(file)
    assert suffix == ".tar.gz"
    os.remove(file)


# TODO: fix for github action
def test_tokencounter_txt_output():
    # TODO: replace with pytest temp dir stuff
    with open("tests/foo.txt", "w") as file:
        cfg = config.Config(
            Path("tests/test_files"),
            True,
            False,
            False,
            False,
            False,
            False,
            Model("gpt-4o"),
            file,
            "txt",
            [],
            [],
        )
        tc = models.TokenCounter(cfg)
        tc.parse_files()
        tc.output()
    with open("tests/output.txt", "r") as f:
        with open("tests/foo.txt", "r") as file:
            assert file.read() == f.read()
    os.remove(Path("tests/foo.txt"))


# TODO: fix for github action
# This covers the TokenCounterEncoder too, technically - it only exists in output()
def test_tokencounter_json_output():
    # TODO: replace with pytest temp dir stuff
    with open("tests/foo.json", "w") as file:
        cfg = config.Config(
            Path("tests/test_files"),
            True,
            False,
            False,
            False,
            False,
            False,
            Model("gpt-4o"),
            file,
            "json",
            [],
            [],
        )
        tc = models.TokenCounter(cfg)
        tc.parse_files()
        tc.output()
    with open("tests/output.json", "r") as f:
        with open("tests/foo.json", "r") as file:
            assert file.read() == f.read()
    os.remove(Path("tests/foo.json"))


# TODO: fix for github action
def test_tokencounter_html_output():
    # TODO: replace with pytest temp dir stuff
    with open("tests/foo.html", "w") as file:
        cfg = config.Config(
            Path("tests/test_files"),
            True,
            False,
            False,
            False,
            False,
            False,
            Model("gpt-4o"),
            file,
            "html",
            [],
            [],
        )
        tc = models.TokenCounter(cfg)
        tc.parse_files()
        tc.output()
    with open("tests/output.html", "r") as f:
        with open("tests/foo.html", "r") as file:
            assert file.read() == f.read()
    os.remove(Path("tests/foo.html"))


"""
    filecategory
"""


def test_filecategory():
    fc = models.FileCategory(".foo")
    assert fc.extension == ".foo"
    assert fc.files == []
    assert fc.total == 0
    assert fc.to_dict() == {"total": 0, "files": []}



"""
Two categories of input:
    - Single Item
        - A single file
        - A stream of text characters
    - Multiple items
        - A list of files


1. Is it a single item?
    - If yes, attempt to parse.
        - If success, pass to Result.Counted
        - If failed, pass to Result.Not Counted
2. If no, spawn threads equivalent to machine capacity:
    2a. Perform Step 1.
"""
